﻿using System;
using System.Collections.Generic;

namespace HeuristicLab.Algorithms.DataAnalysis.SymRegGrammarEnumeration {

  public abstract class Symbol : IEquatable<Symbol> {
    private readonly int stringRepresentationHash;

    public string StringRepresentation { get; }

    protected Symbol(string representation) {
      StringRepresentation = representation;
      stringRepresentationHash = representation.GetHashCode();
    }

    public override string ToString() {
      return StringRepresentation;
    }

    #region IEquatable
    public static bool operator ==(Symbol s1, Symbol s2) {
      if (ReferenceEquals(s1, s2)) return true;
      if (ReferenceEquals(s1, null) || ReferenceEquals(s2, null)) return false;
      return s1.Equals(s2);
    }

    public static bool operator !=(Symbol s1, Symbol s2) {
      return !(s1 == s2);
    }

    public bool Equals(Symbol other) {
      if (ReferenceEquals(other, null)) return false;
      if (ReferenceEquals(other, this)) return true;
      if (this.GetType() != other.GetType()) return false; // Otherwise, this needs to be reimplemented in derived classes.
      return StringRepresentation == other.StringRepresentation;
    }

    public override bool Equals(object obj) {
      if (ReferenceEquals(obj, null)) return false;
      if (ReferenceEquals(obj, this)) return true;
      if (this.GetType() != obj.GetType()) return false;
      return Equals((Symbol)obj);
    }

    public override int GetHashCode() {
      return stringRepresentationHash;
    }
    #endregion
  }

  public class TerminalSymbol : Symbol {

    public TerminalSymbol(string representation) : base(representation) { }
  }

  public class VariableTerminalSymbol : TerminalSymbol {
    public VariableTerminalSymbol(string representation) : base(representation) { }
  }


  public class NonterminalSymbol : Symbol {
    public NonterminalSymbol(string representation) : base(representation) { }
  }

  public class Production : List<Symbol> {

    public Production(params Symbol[] symbols) : base(symbols) { }

    public Production(IEnumerable<Symbol> symbols) : base(symbols) { }

    public override string ToString() {
      return string.Join(" ", this);
    }
  }
}
